/*************************************************************************
 * TLV 02/15/2002                                                        *
 *                                                                       *
 * Ok, I'm not going to claim that this is pretty, but it should work    *
 * to parse fairly error free configuration files.  Since I plan to have *
 * the config files generated by a GUI program at some point, this is    *
 * probably ok.                                                          *
 *                                                                       *
 * Should you care to clean anything up and/or add better error checking *
 * and handling, feel free!                                              *
 *                                                                       *
 * Routines for reading from config files                                *
 *                                                                       *
 *                                                                       *
 * This is free software; you can redistribute it and/or modify          *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation; either version 2 of the License, or     *
 * at your option) any later version.                                    *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with Foobar; if not, write to:                                  *
 *                   Free Software Foundation, Inc.                      *
 *                   59 Temple Place                                     *
 *                   Suite 330                                           *
 *                   Boston, MA  02111-1307                              *
 *                   USA                                                 *
 *************************************************************************/

#include "cfg_util.h"

/************************************************************************
 * Attempt to convert a string containing a hex value to an int         *
 *                                                                      *
 * Valid formats are: 0xyy or yy                                        *
 *  where y can be any of (0-9 and (a-f or A-F))                        *
 *                                                                      *
 *                                                                      *
 * Returns: returns int value on success, else -1                       *
 ************************************************************************/
int htoi(char *hexstr) {
  
   int tempval=0;
   int this_digit =0;
   int pos;

   pos = strlen(hexstr) - 1;

   while (pos >= 0) {
      switch (toupper(*hexstr)) {
      case 'X':
         /*This is the x from 0x - ignore it*/
      case '0':
         /*0 - don't have to do anything*/
	 this_digit = 0;
	 break;
      case '1':
	 this_digit = 1;
	 break;
      case '2':
	 this_digit = 2;
	 break;
      case '3':
	 this_digit = 3;
	 break;
      case '4':
	 this_digit = 4;
	 break;
      case '5':
	 this_digit = 5;
	 break;
      case '6':
	 this_digit = 6;
	 break;
      case '7':
	 this_digit = 7;
	 break;
      case '8':
	 this_digit = 8;
	 break;
      case '9':
	 this_digit = 9;
	 break;
      case 'A':
	 this_digit = 10;
	 break;
      case 'B':
	 this_digit = 11;
	 break;
      case 'C':
	 this_digit = 12;
	 break;
      case 'D':
	 this_digit = 13;
	 break;
      case 'E':
	 this_digit = 14;
	 break;
      case 'F':
	 this_digit = 15;
	 break;
      default:
         return -1;
      }
      if (this_digit) {
         tempval = tempval + (pow(16, pos) * this_digit);
      }
      hexstr++;
      pos--;
   }
   return tempval;
}

/************************************************************************
 * Attempt to read shift code and 64 key codes from config file         *
 *                                                                      *
 * Returns: 0 if read successfully, else -1                             *
 ************************************************************************/
int ReadCfg(char *cfg_file, struct cfg_struct *cfg_data) {

   FILE *stream;
   char buf[255];
   char temp_code[255];
   char temp_label[255];
   char comment[255];
   int temp_num;
   int i;

   if ((stream = fopen(cfg_file, "r")) == NULL) {
      perror("Failed to open config file");
      return -1;
   }

   cfg_data->shift_code = -1;

   /*Now, try to find the Misc section*/
   if (GotoSection(stream, MISC_SECTION)) {
      /*We can read until we find close section*/
      /*Next line should be the shift code*/
      if (ReadNextLine(stream, buf)) {
         /*sscanf(buf,"%s\t0x%s\t%s\n", temp_label, temp_code, comment);*/
         sscanf(buf,"%s\t%s\t%s\n", temp_label, temp_code, comment);
         temp_num = htoi(temp_code);
	 if (strcmp(temp_label, SHIFT_CODE_LABEL) == 0) {
	    cfg_data->shift_code = temp_num;   
	 }
      }
   } else {
      /*Failed to find the misc section*/
      printf("Malformed config file: Failed to find the misc section.\n");
      return -1;
   }

   if (cfg_data->shift_code < 0) {
      printf("Malformed config file: Failed to find shift code.\n");
      return -1;
   }
  

   i=0;
   /*Next, try to find the unshifted section*/
   if (GotoSection(stream, UNSHIFTED_SECTION)) {
      while (ReadNextLine(stream, buf)) {
         sscanf(buf,"%s\t0x%s\t%s\n", temp_label, temp_code, comment);
	 temp_num = htoi(temp_code);
	 cfg_data->key_code[i] = temp_num;
	 i++;
      }
   } else {
      /*Failed to find the unshifted section*/
      printf("Malformed config file: Failed to find the unshifted section.\n");
      return -1;
   }

   if (i<31) {
      printf("Malformed config file: only found %d codes in unshifted config section\n", i);
      return -1;
   }

   /*And finally, try to find the shifted section*/
   if (GotoSection(stream, SHIFTED_SECTION)) {
      while (ReadNextLine(stream, buf)) {
         sscanf(buf,"%s\t0x%s\t%s\n", temp_label, temp_code, comment);
	 temp_num = htoi(temp_code);
	 cfg_data->key_code[i] = temp_num;
	 i++;
      }
   } else {
      /*Failed to find the shifted section*/
      printf("Malformed config file: Failed to find the shifted section.\n");
      return -1;
   }

   if (i<63) {
      printf("Malformed config file: only found %d codes in shifted config section\n", i);
      return -1;
   }


   close(stream);

   /*Normal completion*/
   return 0;
}

/************************************************************************
 * Try to find a section in a config file.  Moves pointer to that point *
 *                                                                      *
 * Returns: 1 if found, else 0                                          *
 ************************************************************************/
int GotoSection(FILE *stream, char *section) {

   char buf[255];
   int i=0;
   int no_seek=0;
   char tempname[255];

   /*Check if we're already at the beginning of the file...*/
   if (ftell(stream) == 0) 
      no_seek = 1;

   /*Look for the section heading*/
   /*Headings look like: [SectionName]*/
   while (!feof(stream)) {
      fgets(buf, 254, stream);
      /*Get rid of any garbage on the end of the string*/
      sscanf(buf, "%s", tempname);
      if (strcmp(tempname, section) == 0) {
         /*Yup, this is the section!*/
         return 1;
      }

      /*Didn't find it -- to verify, seek to the beginning and try again*/
      if ((i=0) && (feof(stream)) && (no_seek != 1)) {
         printf("Moving back to start of file\n");
         rewind(stream);
      }
   }

   /*Didn't find the specified section*/
   return 0;

}


/************************************************************************
 * Try to read the next line in a config file.  Automatically skips     *
 * over comments.                                                       *
 *                                                                      *
 * Returns: 1 if valid, else 0                                          *
 ************************************************************************/
int ReadNextLine(FILE *stream, char *buf_line) {
   char tempbuf[255];
   int done=0;

   while ((!feof(stream)) && (!done)) {
      if (fgets(tempbuf, 254, stream) != NULL) {
         switch (tempbuf[0]) {
	 case '#':
            /*This is a comment line -- skip it*/
	    break;
         case '.':
	    /*Reached end of section -- we can't read any further*/
	    done=1;
	    break;
	 default:
	    /*Found a valid line*/
	    strcpy(buf_line, tempbuf);
	    return 1;
         }
      } else {
         perror("Error reading next line\n");
	 return 0;
      }
   }
   return 0;
}

/************************************************************************
 * Try to generate a default configuration file using the default MAME  *
 * settings.                                                            *
 *                                                                      *
 * Returns: 0 if written ok, else -1                                    *
 ************************************************************************/
int WriteDefaultCfg(char *filename) {

   struct key_struct data_def[64];
   int i;

   for(i=0; i<64; i++) 
      data_def[i] = IPAC_DEFAULTS[i];

   return WriteCfgFile(filename, data_def, DEF_SHIFT);

}

/************************************************************************
 * Try to generate a configuration file                                 *
 *                                                                      *
 * Returns: 0 if written ok, else -1                                    *
 ************************************************************************/
int WriteCfgFile(char *filename, struct key_struct *data_def, 
                 struct key_struct shift_def) {

   FILE *stream;
   int i=0;

   if ((stream = fopen(filename, "w")) == NULL) {
      printf("Failed to open %s for writing\n", filename);
      return -1;
   }

   /*Write out the misc section*/
   fprintf(stream, "%s\n", MISC_SECTION);
   fprintf(stream, "%s\t0x%02X\t;%s - %s\n", SHIFT_CODE_LABEL, 
           shift_def.def_code, shift_def.comment, shift_def.def_key);
   fprintf(stream, "%s\n\n", SECTION_CLOSE);

   /*Write out the unshifted section*/
   fprintf(stream, "%s\n", UNSHIFTED_SECTION);
   for(i=0; i<32; i++) {
      fprintf(stream, "%d\t0x%02X\t;%s - %s\n", i, data_def[i].def_code,
              data_def[i].comment, data_def[i].def_key);
   }
   fprintf(stream, "%s\n\n", SECTION_CLOSE);

   /*Write out the shifted section*/
   fprintf(stream, "%s\n", SHIFTED_SECTION);
   for(i=32; i<64; i++) {
      fprintf(stream, "%d\t0x%02X\t;%s - %s\n", (i - 32), data_def[i].def_code,
              data_def[i].comment, data_def[i].def_key);
   }
   fprintf(stream, "%s\n\n", SECTION_CLOSE);
   
   fflush(stream);
   close(stream);

   /*Normal completion*/
   return 0;
}
